import pytest
from n2y.plugins.jinjarenderpage import FirstPassOutput


class TestFirstPassOutput:
    def test_create(self):
        first_pass_output = FirstPassOutput()
        assert first_pass_output is not None
        assert not first_pass_output

    def test_first_pass_behavior(self):
        first_pass_output = FirstPassOutput()
        assert not first_pass_output.is_second_pass
        assert not first_pass_output.second_pass_is_requested
        assert first_pass_output.lines == []
        assert first_pass_output.second_pass_is_requested
        assert first_pass_output.lines == []
        assert first_pass_output.source == ''

    @pytest.mark.parametrize('source_lines, expected_source',
                             [
                                 (
                                     [],
                                     '',
                                 ),
                                 (
                                     ['alpha Beta', 'GAMMA', 'beta'],
                                     'alpha Beta\nGAMMA\nbeta',
                                 ),
                             ])
    def test_second_pass_behavior(self, source_lines, expected_source):
        first_pass_output = FirstPassOutput(source_lines)
        assert first_pass_output.is_second_pass
        assert not first_pass_output.second_pass_is_requested
        assert first_pass_output.lines == source_lines
        assert first_pass_output.source == expected_source
        assert not first_pass_output.second_pass_is_requested



"""
These tests verify how the n2y block classes convert notion data into Pandoc
abstract syntax tree (AST) objects, and then into markdown.
"""
from pandoc.types import (
    Str, Space, Header, Link
)

from n2y.utils import strip_hyphens
from n2y.notion_mocks import mock_block, mock_rich_text
from tests.test_blocks import process_block


def mock_header_ast(level, suffix, notion_block):
    return Header(
        level,
        ("", [], []),
        [
            Link(
                ("", [], []),
                [Str("Heading"), Space(), Str(suffix)],
                (f'#{strip_hyphens(notion_block["id"])}', ""),
            )
        ],
    )


def test_heading_1():
    notion_block = mock_block("heading_1", {"rich_text": [mock_rich_text("Heading One")]})
    pandoc_ast, markdown = process_block(notion_block, ["n2y.plugins.linkedheaders"])
    assert pandoc_ast == mock_header_ast(1, "One", notion_block)
    assert markdown == f'# [Heading One]({notion_block["url"]})\n'


def test_heading_1_bolding_stripped():
    notion_block = mock_block("heading_1", {"rich_text": [mock_rich_text("Heading One", ["bold"])]})
    pandoc_ast, markdown = process_block(notion_block, ["n2y.plugins.linkedheaders"])
    assert pandoc_ast == mock_header_ast(1, "One", notion_block)
    assert markdown == f'# [Heading One]({notion_block["url"]})\n'


def test_heading_2():
    notion_block = mock_block("heading_2", {"rich_text": [mock_rich_text("Heading Two")]})
    pandoc_ast, markdown = process_block(notion_block, ["n2y.plugins.linkedheaders"])
    assert pandoc_ast == mock_header_ast(2, "Two", notion_block)
    assert markdown == f'## [Heading Two]({notion_block["url"]})\n'


def test_heading_3():
    notion_block = mock_block("heading_3", {"rich_text": [mock_rich_text("Heading Three")]})
    pandoc_ast, markdown = process_block(notion_block, ["n2y.plugins.linkedheaders"])
    assert pandoc_ast == mock_header_ast(3, "Three", notion_block)
    assert markdown == f'### [Heading Three]({notion_block["url"]})\n'



from unittest.mock import patch, PropertyMock
from jinja2 import TemplateSyntaxError
from pytest import raises

from n2y.page import Page
from n2y.notion import Client
from n2y.blocks import ChildPageBlock
from n2y.utils import pandoc_ast_to_markdown, pandoc_write_or_log_errors
from n2y.plugins.jinjarenderpage import (
    render_from_string, join_to, fuzzy_find_in,
    JinjaFencedCodeBlock, JinjaRenderPage,
)
from n2y.notion_mocks import (
    mock_database, mock_database_mention, mock_page, mock_rich_text,
    mock_rich_text_array, mock_block,
)


def test_join_to_basic():
    foreign_keys = ['1', '3']
    table = [
        {'id': '1', 'data': 'a'},
        {'id': '2', 'data': 'b'},
    ]
    assert join_to(foreign_keys, table) == [{'id': '1', 'data': 'a'}, None]
    assert join_to(foreign_keys, table, 'data') == [None, None]


def test_fuzzy_find_in():
    dict_list = [
        {'id': '1', 'data': 'a'},
        {'id': '2', 'data': 'bc'},
        {'id': '3', 'data': 'def'},
    ]
    reversed_dict_list = [
        {'id': '3', 'data': 'def'},
        {'id': '2', 'data': 'bc'},
        {'id': '1', 'data': 'a'}
    ]
    catch_all_string = 'a bc def'
    assert fuzzy_find_in(dict_list, 'a', 'data') == [{'id': '1', 'data': 'a'}]
    assert fuzzy_find_in(dict_list, catch_all_string, 'data') == reversed_dict_list
    assert fuzzy_find_in(dict_list, catch_all_string, 'data', False) == reversed_dict_list
    assert fuzzy_find_in(dict_list, catch_all_string, 'data', False, False) == dict_list
    assert fuzzy_find_in(dict_list, catch_all_string, 'data', True, False) == dict_list


def test_render_no_filtering():
    input_string = "apple\nbanana\ncherry\n"
    expected_result = input_string
    actual_result = render_from_string(input_string)
    assert actual_result == expected_result


def test_undefined():
    with raises(TemplateSyntaxError):
        input_string = "{% huhwhat 'hotel', 'california' %}"
        render_from_string(input_string)


def process_jinja_block(client, caption, jinja_code):
    page_notion_data = mock_page()
    page = JinjaRenderPage(client, page_notion_data)

    page_block_notion_data = mock_block('child_page', {'title': "Mock Page"})
    page_block = ChildPageBlock(client, page_block_notion_data, page, False)

    page._block = page_block

    code_block_notion_data = mock_block('code', {
        'language': 'plain text',
        'caption': caption,
        'rich_text': mock_rich_text_array(jinja_code)
    })
    code_block = JinjaFencedCodeBlock(client, code_block_notion_data, page, False)
    page.block.children = [code_block]
    return page


def test_jinja_syntax_err():
    client = Client('')
    caption = mock_rich_text_array('{jinja=gfm}')
    jinja_code = "{% huhwhat 'hotel', 'california' %}"
    page = process_jinja_block(client, caption, jinja_code)
    with raises(TemplateSyntaxError):
        page.to_pandoc()


def test_jinja_render_gfm():
    client = Client('')
    caption = mock_rich_text_array('{jinja=gfm}')
    jinja_code = "{% for v in ['a', 'b'] %}{{v}}{% endfor %}"
    page = process_jinja_block(client, caption, jinja_code)
    pandoc_ast = page.to_pandoc()
    markdown = pandoc_ast_to_markdown(pandoc_ast)
    assert markdown == "ab\n"


def test_jinja_render_gfm_with_second_pass():
    client = Client('')
    caption = mock_rich_text_array('{jinja=gfm}')
    jinja_code = "a{% for v in first_pass_output.lines %}{{v}}{% endfor %}"
    page = process_jinja_block(client, caption, jinja_code)
    page_block = page.block
    with patch.object(Page, "block", new_callable=PropertyMock) as mock_block:
        mock_block.return_value = page_block
        pandoc_ast = page.to_pandoc()
    markdown = pandoc_ast_to_markdown(pandoc_ast)
    assert markdown == "aa\n"


def test_jinja_render_html():
    client = Client('')
    caption = mock_rich_text_array('{jinja=html}')
    jinja_code = (
        "<table>"
        "<tr><th>Name</th></tr>"
        "{% for v in ['a', 'b'] %}<tr><td>{{v}}</td></tr>{% endfor %}"
        "</table>"
    )
    page = process_jinja_block(client, caption, jinja_code)
    pandoc_ast = page.to_pandoc()
    markdown = pandoc_ast_to_markdown(pandoc_ast)
    assert markdown == (
        "  Name\n"
        "  ------\n"
        "  a\n"
        "  b\n"
    )


def test_jinja_render_with_database():
    client = Client('')
    database_notion_data = mock_database(title='My DB')
    mention_notion_data = mock_database_mention(database_notion_data['id'])
    database_pages_notion_data = [mock_page(title='a'), mock_page(title='b')]
    caption = [
        mock_rich_text('{jinja=gfm} '),
        mock_rich_text('My DB', mention=mention_notion_data),
    ]
    jinja_code = "{% for v in databases['My DB'] %}{{v.title}}{% endfor %}"

    with patch.object(Client, "get_notion_database") as mock_get_notion_database:
        with patch.object(Client, "get_database_notion_pages") as mock_get_database_notion_pages:
            mock_get_notion_database.return_value = database_notion_data
            mock_get_database_notion_pages.return_value = database_pages_notion_data
            page = process_jinja_block(client, caption, jinja_code)
            pandoc_ast = page.to_pandoc()

    markdown = pandoc_ast_to_markdown(pandoc_ast)
    assert markdown == "ab\n"


def test_jinja_render_plain():
    client = Client('')
    caption = mock_rich_text_array('{jinja=plain}')
    jinja_code = "# <h1> asdf23"
    page = process_jinja_block(client, caption, jinja_code)
    pandoc_ast = page.to_pandoc()
    text = pandoc_write_or_log_errors(pandoc_ast, 'plain', [])
    assert text == "# <h1> asdf23\n"



import copy
import yaml

from n2y.config import (
    valid_notion_id, merge_config, load_config, _valid_notion_filter,
    _validate_config_item, EXPORT_DEFAULTS
)
from n2y.notion_mocks import mock_id


def mock_config_item(node_type):
    config_item = copy.deepcopy(EXPORT_DEFAULTS)
    config_item["id"] = mock_id()
    config_item["node_type"] = node_type
    return config_item


def test_load_config_basic(tmp_path):
    # use a temporary file to test the config loading
    config_path = tmp_path / "config.yaml"
    export_id = mock_id()
    with open(config_path, "w") as f:
        f.write(yaml.dump({
            "media_root": "media",
            "media_url": "https://example.com/media",
            "export_defaults": {
                "id_property": "id",
                "url_property": "url",
            },
            "exports": [
                {
                    "id": export_id,
                    "node_type": "page",
                    "output": "output.md",
                    "pandoc_format": "gfm",
                }
            ]
        }))
    config = load_config(config_path)
    assert config is not None, "The config is invalid"
    merged_export = config["exports"][0]
    assert merged_export["id"] == export_id
    assert merged_export["node_type"] == "page"
    assert merged_export["id_property"] == "id"
    assert merged_export["url_property"] == "url"
    assert merged_export["pandoc_format"] == "gfm"


def test_merge_config_no_defaults():
    master_defaults = {"a": "1"}
    defaults = {}
    config_items = [
        {"b": "1"},
        {"b": "2", "a": "2"},
    ]
    assert merge_config(config_items, master_defaults, defaults) == [
        {"a": "1", "b": "1"},
        {"b": "2", "a": "2"},
    ]


def test_merge_config_defaults():
    master_defaults = {"a": "1", "b": "1"}
    defaults = {"a": "3"}
    config_items = [
        {},
        {"a": "2"},
        {"b": "2"},
    ]
    assert merge_config(config_items, master_defaults, defaults) == [
        {"a": "3", "b": "1"},
        {"a": "2", "b": "1"},
        {"a": "3", "b": "2"},
    ]


def test_valid_id_valid():
    assert valid_notion_id(mock_id())


def test_valid_id_invalid():
    assert not valid_notion_id(mock_id() + 'a')


def test_valid_id_invalid_due_to_special():
    bad_id = 'https://' + mock_id()[8:]
    assert not valid_notion_id(bad_id)


def test_valid_notion_filter_simple():
    assert _valid_notion_filter({
        "property": "title",
        "direction": "ascending",
    })


def test_valid_notion_filter_complex():
    assert _valid_notion_filter([{
        "property": "title",
        "direction": "ascending",
    }])


def test_valid_config_item_missing_id():
    config_item = mock_config_item("page")
    del config_item["id"]
    assert not _validate_config_item(config_item)


def test_valid_config_item_missing_node_type():
    config_item = mock_config_item("page")
    del config_item["node_type"]
    assert not _validate_config_item(config_item)


def test_valid_config_item_invalid_node_type():
    config_item = mock_config_item("page")
    config_item["node_type"] = "invalid"
    assert not _validate_config_item(config_item)


def test_valid_config_item_missing_filename_template():
    config_item = mock_config_item("database_as_files")
    assert not _validate_config_item(config_item)


def test_valid_config_item_malformed_filename_template():
    config_item = mock_config_item("database_as_files")
    config_item["filename_template"] = "{"
    assert not _validate_config_item(config_item)



from pandoc.types import (
    Str, Space, Strong, Emph, Strikeout, Code, Link, Math, InlineMath,
    Underline
)

from n2y.notion import Client
from n2y.rich_text import RichTextArray, MentionRichText
from n2y.mentions import PageMention
from n2y.notion_mocks import mock_rich_text_array, mock_annotations, mock_rich_text, mock_id


def process_rich_text_array(notion_data):
    client = Client('')
    rich_text_array = client.wrap_notion_rich_text_array(notion_data)
    pandoc_ast = rich_text_array.to_pandoc()
    markdown = rich_text_array.to_value('markdown')
    plain_text = rich_text_array.to_plain_text()
    return pandoc_ast, markdown, plain_text


def test_bold_word():
    notion_data = mock_rich_text_array([('A ', []), ('bold', ['bold']), (' word.', [])])
    pandoc_ast, markdown, plain_text = process_rich_text_array(notion_data)
    assert pandoc_ast == [
        Str("A"), Space(), Strong([Str("bold")]), Space(), Str("word.")
    ]
    assert markdown == "A **bold** word.\n"
    assert plain_text == "A bold word."


def test_bold_letter():
    notion_data = mock_rich_text_array([('A ', []), ('b', ['bold']), ('old word.', [])])
    pandoc_ast, markdown, plain_text = process_rich_text_array(notion_data)
    assert pandoc_ast == [
        Str("A"), Space(), Strong([Str("b")]), Str("old"), Space(), Str("word.")
    ]
    assert markdown == "A **b**old word.\n"
    assert plain_text == "A bold word."


def test_bold_spaces():
    notion_data = mock_rich_text_array([('A', []), (' bold ', ['bold']), ('word.', [])])
    pandoc_ast, markdown, plain_text = process_rich_text_array(notion_data)
    assert pandoc_ast == [
        Str("A"), Space(), Strong([Str("bold")]), Space(), Str("word.")
    ]
    assert markdown == "A **bold** word.\n"
    assert plain_text == "A bold word."


def test_italic_word():
    notion_data = mock_rich_text_array([('An ', []), ('italic', ['italic']), (' word.', [])])
    pandoc_ast, markdown, _ = process_rich_text_array(notion_data)
    assert pandoc_ast == [
        Str("An"), Space(), Emph([Str("italic")]), Space(), Str("word.")
    ]
    assert markdown == "An *italic* word.\n"


def test_italic_letter():
    notion_data = mock_rich_text_array([('An ', []), ('i', ['italic']), ('talic word.', [])])
    pandoc_ast, markdown, _ = process_rich_text_array(notion_data)
    assert pandoc_ast == [Str("An"), Space(), Emph([Str("i")]), Str("talic"), Space(), Str("word.")]
    assert markdown == "An *i*talic word.\n"


def test_bold_italic_word():
    notion_data = mock_rich_text_array([
        ('A ', []),
        ('bold-italic', ['bold', 'italic']),
        (' word.', []),
    ])
    pandoc_ast, markdown, plain_text = process_rich_text_array(notion_data)
    assert pandoc_ast == [
        Str('A'),
        Space(),
        Emph([Strong([Str('bold-italic')])]),
        Space(),
        Str('word.'),
    ]
    assert markdown == "A ***bold-italic*** word.\n"
    assert plain_text == "A bold-italic word."


def test_bold_space():
    notion_data = mock_rich_text_array([
        (' ', ['bold']),
    ])
    pandoc_ast, _, _ = process_rich_text_array(notion_data)
    assert pandoc_ast == [Space()]


def test_italic_spaces():
    notion_data = mock_rich_text_array([('An', []), (' italic ', ['italic']), ('word.', [])])
    pandoc_ast, markdown, plain_text = process_rich_text_array(notion_data)
    assert pandoc_ast == [Str('An'), Space(), Emph([Str('italic')]), Space(), Str('word.')]
    assert markdown == "An *italic* word.\n"
    assert plain_text == "An italic word."


def test_strikeout_word():
    notion_data = mock_rich_text_array([('A ', []), ('deleted', ['strikethrough']), (' word.', [])])
    pandoc_ast, markdown, plain_text = process_rich_text_array(notion_data)
    assert pandoc_ast == [Str("A"), Space(), Strikeout([Str("deleted")]), Space(), Str("word.")]
    assert markdown == "A ~~deleted~~ word.\n"
    assert plain_text == "A deleted word."


def test_strikeout_spaces():
    notion_data = mock_rich_text_array([('A', []), (' deleted ', ['strikethrough']), ('word.', [])])
    pandoc_ast, markdown, _ = process_rich_text_array(notion_data)
    assert pandoc_ast == [Str("A"), Space(), Strikeout([Str("deleted")]), Space(), Str("word.")]
    assert markdown == "A ~~deleted~~ word.\n"


def test_blended_annotated_spaces():
    notion_data = mock_rich_text_array([
        ('this ', ['bold']),
        ('is ', ['bold', 'italic']),
        ('a', ['italic']),
        (' test', ['strikethrough']),
        (' did', ['bold', 'underline', 'code']),
        (' i pass', ['underline', 'code']),
        ('?', [])
    ])
    pandoc_ast, markdown, _ = process_rich_text_array(notion_data)
    assert pandoc_ast == [
        Strong([Str('this')]),
        Space(),
        Emph([Strong([Str('is')])]),
        Space(),
        Emph([Str('a')]),
        Space(),
        Strikeout([Str('test')]),
        Underline([Strong([Code(('', [], []), ' did')])]),
        Underline([Code(('', [], []), ' i pass')]),
        Str('?')
    ]
    assert markdown == (
        '**this** ***is*** *a*\n~~test~~[**` did`*'
        '*]{.underline}[` i pass`]{.underline}?\n'
    )


def test_equation_inline():
    lhs = "{\\displaystyle i\\hbar {\\frac {d}{dt}}\\vert "
    rhs = "\\Psi (t)\\rangle={\\hat {H}}\\vert \\Psi (t)\\rangle}"
    example_equation = f"{lhs}{rhs}"

    notion_data = [
        mock_rich_text('The Schrödinger Equation ('),
        {
            'type': 'equation',
            'equation': {'expression': example_equation},
            'annotations': mock_annotations(),
            'plain_text': example_equation,
            'href': None,
        },
        mock_rich_text(') is a very useful one indeed'),
    ]

    pandoc_ast, markdown, _ = process_rich_text_array(notion_data)
    assert pandoc_ast == [
        Str('The'), Space(), Str('Schrödinger'), Space(), Str('Equation'), Space(), Str('('),
        Math(InlineMath(), example_equation), Str(')'), Space(),
        Str('is'), Space(), Str('a'), Space(), Str('very'), Space(),
        Str('useful'), Space(), Str('one'), Space(), Str('indeed'),
    ]
    assert markdown == (
        'The Schrödinger Equation\n(${\\displaystyle '
        'i\\hbar {\\frac {d}{dt}}\\vert \\Psi (t)\\'
        'rangle={\\hat {H}}\\vert \\Psi (t)\\rangle}$)\nis'
        ' a very useful one indeed\n'
    )


def test_code_inline():
    notion_data = mock_rich_text_array([('A ', []), ('code', ['code']), (' word.', [])])
    pandoc_ast, markdown, plain_text = process_rich_text_array(notion_data)
    assert pandoc_ast == [Str("A"), Space(), Code(("", [], []), "code"), Space(), Str("word.")]
    assert markdown == "A `code` word.\n"
    assert plain_text == "A code word."


def test_link_inline():
    notion_data = [
        mock_rich_text("This is a "),
        mock_rich_text("link", ["bold"], href="https://example.com"),
        mock_rich_text("."),
    ]
    pandoc_ast, markdown, plain_text = process_rich_text_array(notion_data)
    assert pandoc_ast == [
        Str('This'), Space(), Str('is'), Space(), Str('a'), Space(),
        Link(('', [], []), [Strong([Str('link')])], ('https://example.com', '')),
        Str('.'),
    ]
    assert markdown == 'This is a [**link**](https://example.com).\n'
    assert plain_text == 'This is a link.'


def test_prepend():
    client = Client('')
    plain_data = mock_rich_text_array('plain text')
    mention_data = {'type': 'page', 'page': {'id': mock_id()}}
    mention_text = 'Test Page'
    mention_block = PageMention(client, mention_data, mention_text)
    mention_rich_text = MentionRichText(
        client,
        mock_rich_text(
            mention_text,
            None,
            None,
            mention_data
        ),
        None,
        mention_block
    )
    equation_text = 'e=mc^2'
    equation_data = mock_rich_text_array([[
        equation_text,
        None,
        None,
        None,
        None,
        {'expression': equation_text}
    ]])
    plain = RichTextArray(client, plain_data)
    mention = RichTextArray(client, [])
    equation = RichTextArray(client, equation_data)
    mention.items.append(mention_rich_text)
    assert plain.to_pandoc() == [Str('plain'), Space(), Str('text')]
    assert mention.to_pandoc() == [Str('Test'), Space(), Str('Page')]
    assert equation.to_pandoc() == [Math(InlineMath(), equation_text)]



from math import isclose
from pytest import raises
from datetime import datetime, timezone, timedelta

import pytest
from pandoc.types import MetaMap, MetaList, MetaBool, MetaString

from n2y.errors import APIResponseError
from n2y.utils import (
    fromisoformat, header_id_from_text, id_from_share_link, retry_api_call,
    yaml_to_meta_value,
)


def test_fromisoformat_datetime():
    expected = datetime(2022, 5, 10, 19, 52, tzinfo=timezone.utc)
    assert fromisoformat('2022-05-10T19:52:00.000Z') == expected


def test_fromisoformat_date():
    expected = datetime(2022, 5, 11)
    assert fromisoformat('2022-05-11') == expected


def test_database_id_from_share_link_id():
    database_id = "f77db3af4ab6a47b6162dacd76681231"
    assert id_from_share_link(database_id) == database_id


def test_database_id_from_share_link_id_hyphens():
    database_id = "f77db3af4ab6a47b6162dacd76681231"
    database_id_hyphens = "f77db3af-4ab6a47b6162dacd-76681231"
    assert id_from_share_link(database_id_hyphens) == database_id


def test_database_id_from_share_link_no_hyphens():
    database_id = "f77db3af4ab6a47b6162dacd76681231"
    view_id = "463a296148dc38f791e7037dda9a8c3f"
    share_link = f"https://www.notion.so/ws/{database_id}?v={view_id}"
    assert id_from_share_link(share_link) == database_id


def test_database_id_from_share_link_hyphens():
    database_id = "90a3f77db3af4ab6a47b6162dacd1111"
    database_id_hyphens = "90a3f77-db3af4ab6a47b6162-dacd1111"
    view_id = "463a296148dc38f791e7037dda9a8c3f"
    share_link = f"https://www.notion.so/ws/{database_id_hyphens}?v={view_id}"
    assert id_from_share_link(share_link) == database_id


def test_page_id_from_share_link():
    page_id = '42361d2334624e949d9b7e94d30c988b'
    share_link = f'https://www.notion.so/innolitics/SaMD-DHF-Template-{page_id}'
    assert id_from_share_link(share_link) == page_id


class MockResponse():
    def __init__(self, time, code):
        self.headers = {'retry-after': time}
        self.text = ''
        self.status_code = code


def test_retry_api_call_no_error():
    @retry_api_call
    def tester():
        return True
    assert tester()


def test_retry_api_call_multiple_errors():
    status_code = 429

    call_count = 0

    @retry_api_call
    def tester(time):
        seconds = timedelta.total_seconds(datetime.now() - time)
        nonlocal call_count
        call_count += 1

        if call_count == 1:
            raise APIResponseError(MockResponse(0.05, status_code), '', status_code)
        elif call_count == 2:
            assert isclose(0.05, seconds, abs_tol=0.1)
            raise APIResponseError(MockResponse(0.23, status_code), '', status_code)
        elif call_count == 3:
            assert isclose(0.35, seconds, abs_tol=0.1)
            raise APIResponseError(MockResponse(0.16, status_code), '', status_code)
        elif call_count == 4:
            assert isclose(0.51, seconds, abs_tol=0.1)
            return True

    assert tester(datetime.now())


@pytest.mark.parametrize("status_code", [409, 429, 500, 502, 504])
def test_retry_api_call_onc(status_code):
    call_count = 0

    @retry_api_call
    def tester():
        nonlocal call_count
        call_count += 1

        if call_count == 1:
            raise APIResponseError(MockResponse(0.001, status_code), '', status_code)
        else:
            return True

    assert tester()
    assert call_count == 2


def test_retry_api_call_max_errors():
    status_code = 429

    @retry_api_call
    def tester():
        raise APIResponseError(MockResponse(0.001, status_code), '', status_code)
    with raises(APIResponseError):
        tester()


def test_yaml_to_meta_value_scalar():
    assert yaml_to_meta_value('test') == MetaString('test')
    assert yaml_to_meta_value(3) == MetaString('3')
    assert yaml_to_meta_value(3.45) == MetaString('3.45')
    assert yaml_to_meta_value(True) == MetaBool(True)
    assert yaml_to_meta_value(False) == MetaBool(False)
    assert yaml_to_meta_value(None) == MetaString('')


def test_yaml_to_meta_value_map():
    assert yaml_to_meta_value({'a': '1', 'b': 2}) == MetaMap({
        'a': MetaString('1'),
        'b': MetaString('2'),
    })


def test_yaml_to_meta_value_list():
    assert yaml_to_meta_value(['a', 1]) == MetaList([MetaString('a'), MetaString('1')])


def test_header_id_from_text_basic():
    assert header_id_from_text("Essays") == 'essays'
    assert header_id_from_text("Hello Goodbye") == 'hello-goodbye'


def test_header_id_from_text():
    """
    Most of these test cases were taken from the pandoc documentation. See:
    https://pandoc.org/MANUAL.html#extension-auto_identifiers
    """
    assert header_id_from_text('Heading identifiers in HTML') == 'heading-identifiers-in-html'
    assert header_id_from_text("Maître d'hôtel") == 'maître-dhôtel'
    assert header_id_from_text('*Dogs*?--in *my* house?') == 'dogs--in-my-house'
    assert header_id_from_text('[HTML], [S5], or [RTF]?') == 'html-s5-or-rtf'
    assert header_id_from_text('3. Applications') == 'applications'
    assert header_id_from_text('33') == 'section'

    # These are additional tests constructed by us.
    assert header_id_from_text('33 section') == 'section'
    assert header_id_from_text('') == 'section'
    assert header_id_from_text('newlines\nare hyphens') == 'newlines-are-hyphens'


def test_header_id_from_text_existing_ids():
    assert header_id_from_text('a', {'a'}) == 'a-1'
    assert header_id_from_text('a', {'a', 'a-1'}) == 'a-2'
    assert header_id_from_text('', {'section', 'section-1'}) == 'section-2'
    assert header_id_from_text('', {'section', 'a', 'a-1'}) == 'section-1'



"""
The JinjaRenderPage plugin makes it possible to use Jinja2 templates to render
content from Notion databases into the Pandoc AST.

In addition to the raw power of Jinja2, this plugin provides a few useful Jinja
filters and context variables. One variable in particular, the FirstPassOutput
object, is used to make it possible to render content in two passes. This is
useful for rendering content that depends on other content in the same page. For
example, if one ones to generate a glossary from a centralized "Terms" database
and to only include terms which are present on the page, one can use the
FirstPassOutput filter to first generate the content of the page, and then make
a second pass to render the glossary.

Since the FirstPassOutput object requires a full render of the page, this plugin
overrides the Page class.

In addition, it also overrides the CodeBlock class.

Any code block whose caption begins with "{jinja=pandoc_format}" will be rendered
using Jinja2 into text, which in turn will be parsed by Pandoc into assuming it
follows the pandoc_format, into the AST. Thus, the jinja can be used to render
latex, html, or any other format that Pandoc supports into the AST.

Another context variable that is provided is the database variable, which is
used to access Notion databases. Any database that is @-mentioned in the caption
of the codeblock may be accessed via the database name.
"""
import re
import logging

import pandoc
from pandoc.types import Plain, Str
import jinja2
from pandoc.types import Pandoc, Meta, MetaString
from jinja2.exceptions import TemplateSyntaxError, UndefinedError

from n2y.blocks import FencedCodeBlock, HeadingBlock
from n2y.errors import UseNextClass
from n2y.mentions import DatabaseMention
from n2y.page import Page
from n2y.rich_text import MentionRichText
from n2y.utils import pandoc_ast_to_markdown
from n2y.export import database_to_yaml


logger = logging.getLogger(__name__)


class JinjaDatabaseCache(dict):
    def __getitem__(self, key):
        if isinstance(key, str):
            return super().__getitem__(key)
        elif isinstance(key, int):
            return super().__getitem__(list(self)[key])
        logger.error((
            'Jinja template database must be '
            'selected using a string or an integer'
        ))
        raise KeyError(key)


class FirstPassOutput:
    def __init__(self, lines=None):
        self._second_pass_is_requested = False
        self._lines = lines
        self._source = None

    def __bool__(self):
        return bool(self._lines)

    def request_second_pass(self):
        self._second_pass_is_requested = True

    @property
    def is_second_pass(self):
        return self._lines is not None

    @property
    def second_pass_is_requested(self):
        return not self.is_second_pass and self._second_pass_is_requested

    @property
    def lines(self):
        self.is_second_pass or self.request_second_pass()
        return self._lines or []

    def set_lines(self, lines):
        self._lines = lines

    @property
    def source(self):
        if self._source is None or self.is_second_pass:
            self._source = '\n'.join(self.lines)
        return self._source


def join_to(foreign_keys, table, primary_key='id'):
    '''
    Given a set of ids for an object, and a list of the objects these ids refer
    to, select out the objects by joining using the specified primary key
    (which defaults to 'id').
    '''
    joined = []
    for foreign_key in foreign_keys:
        selected_row = None
        for row in table:
            if row[primary_key] == foreign_key:
                selected_row = row
                break
        joined.append(selected_row)
    return joined


def list_matches(string, text):
    return list(re.finditer(
        '(?<![a-zA-Z])' + _canonicalize(string) + '(?![a-zA-Z])',
        _canonicalize(text))
    )


def remove_words(words, text):
    for word in words:
        span = word.span()
        text = text[:span[0]] + ' ' * (span[1] - span[0]) + text[span[1]:]
    return text


def _fuzzy_find_in(term_list, text, key='Name', by_length=True, reverse=True):
    found = []
    key_filter = lambda d: len(d[key]) if by_length else d[key]
    sorted_term_list = sorted(term_list, key=key_filter, reverse=reverse)
    if key in term_list[0]:
        for term in sorted_term_list:
            if key in term and term[key] != '':
                matches = list_matches(term[key], text)
                if matches != []:
                    found.append(term)
                    text = remove_words(matches, text)
    return found


def fuzzy_find_in(term_list, text, key='Name', by_length=True, reverse=True):
    """
    Used to search a markdown string, which may have been modified using pandoc's smart extension,
    for the value at a specified key in a dicionary for all dictionaries in a given list of them.

    see https://pandoc.org/MANUAL.html#extension-smart
    """
    found = []
    if isinstance(key, str):
        found = _fuzzy_find_in(term_list, text, key, by_length, reverse)
    elif isinstance(key, list):
        keys = key
        for key in keys:
            terms_found = fuzzy_find_in(term_list, text, key, by_length, reverse)
            found.extend(terms_found)
    return found


def _canonicalize(markdown):
    markdown = markdown.replace('\u201D', '"').replace('\u201C', '"')
    markdown = markdown.replace('\u2019', "'").replace('\u2018', "'")
    markdown = markdown.replace('\u2013', '--').replace('\u2014', '---')
    markdown = markdown.replace('\u2026', '...')
    markdown = markdown.replace('\u00A0', ' ')
    return markdown


def _create_jinja_environment():
    environment = jinja2.Environment(
        cache_size=0,
        undefined=jinja2.StrictUndefined,
        extensions=["jinja2.ext.do"],
    )
    environment.globals['first_pass_output'] = FirstPassOutput()
    environment.filters['fuzzy_find_in'] = fuzzy_find_in
    environment.filters['join_to'] = join_to
    return environment


def render_from_string(source, context=None, environment=None):
    if environment is None:
        environment = _create_jinja_environment()
    if context is None:
        context = {}
    template = environment.from_string(source)
    output = template.render(context)

    # output string usually loses trailing new line.
    if output and output[-1] != '\n':
        output += '\n'
    return output


class JinjaRenderPage(Page):
    def __init__(self, client, notion_data):
        super().__init__(client, notion_data)
        if 'jinjarenderpage' not in client.plugin_data:
            client.plugin_data['jinjarenderpage'] = {}
        if self.notion_id not in client.plugin_data:
            client.plugin_data['jinjarenderpage'][self.notion_id] = {
                'environment': _create_jinja_environment(),
            }

    def to_pandoc(self):
        first_pass_ast = super().to_pandoc()
        jinja_environment = self.client.plugin_data[
            'jinjarenderpage'][self.notion_id]['environment']
        first_pass_output = jinja_environment.globals["first_pass_output"]
        if first_pass_output.second_pass_is_requested:
            first_pass_output_text = pandoc_ast_to_markdown(first_pass_ast)
            first_pass_output.set_lines(first_pass_output_text.splitlines(keepends=True))
            super().__init__(self.client, self.notion_data)
            second_pass_ast = super().to_pandoc()
            jinja2.clear_caches()
            return second_pass_ast
        else:
            return first_pass_ast


class JinjaFencedCodeBlock(FencedCodeBlock):
    trigger_regex = re.compile(r'^{jinja=(.+)}')

    def __init__(self, client, notion_data, page, get_children=True):
        super().__init__(client, notion_data, page, get_children)
        result = self.caption.matches(self.trigger_regex)
        if result:
            self.pandoc_format = result.group(1)
        else:
            raise UseNextClass()
        self.rendered_text = ''
        self.databases = JinjaDatabaseCache()

    def _get_database_ids_from_mentions(self):
        for rich_text in self.caption:
            is_mention = isinstance(rich_text, MentionRichText)
            if is_mention and isinstance(rich_text.mention, DatabaseMention):
                yield rich_text.mention.notion_database_id

    def _get_yaml_from_mentions(self):
        export_defaults = self.client.export_defaults
        for database_id in self._get_database_ids_from_mentions():
            database = self.client.get_database(database_id)
            # TODO: Rethink about the database data is accessed from within the
            # templates; perhaps it should be something more like Django's ORM
            # where we can filter and sort the databases via the API, instead of
            # having to pull in ALL of the data first before filtering it.
            # Such an API would also alleviate the need to pass in the export
            # defaults and it could provide a mechanism to access the page
            # content from within the Jinja templates, which isn't possible
            # right now.
            database_name = database.title.to_plain_text()
            if database_name in self.databases:
                msg = (
                    f'Duplicate database name "{database_name}"'
                    f' when rendering [{self.notion_url}]'
                )
                logger.error(msg)
                raise ValueError(msg)
            self.databases[database_name] = database_to_yaml(
                database=database,
                pandoc_format=self.pandoc_format,
                pandoc_options=[],
                id_property=export_defaults["id_property"],
                url_property=export_defaults["url_property"],
            )

    def _render_text(self):
        jinja_environment = self.client.plugin_data[
            'jinjarenderpage'][self.page.notion_id]['environment']
        jinja_code = self.rich_text.to_plain_text()
        context = {
            "databases": self.databases,
            "page": self.page.properties_to_values(self.pandoc_format),
        }
        if 'render_content' in jinja_code:
            def render_content(notion_id, level_adjustment=0):
                page = self.client.get_page(notion_id)
                for child in page.block.children:
                    if isinstance(child, HeadingBlock):
                        child.level = max(1, child.level + level_adjustment)
                ast = page.to_pandoc() if page.to_pandoc() else \
                    Pandoc(Meta({'title': MetaString(page.block.title)}), [])
                content = pandoc.write(
                    ast,
                    format=self.pandoc_format,
                    options=[],
                )
                return content
            self.client.plugin_data['jinjarenderpage'][self.page.notion_id][
                'environment'].filters['render_content'] = render_content
        try:
            self.rendered_text = render_from_string(jinja_code, context, jinja_environment)
        except (UndefinedError, TemplateSyntaxError):
            logger.error(
                "Error rendering Jinja template on %s (%r)",
                self.page.title.to_plain_text() if self.page else "unknown",
                self.notion_url,
                exc_info=True,
            )
            raise

    def to_pandoc(self):
        if not len(self.databases):
            self._get_yaml_from_mentions()
            self._render_text()
        if self.pandoc_format != "plain":
            # pandoc.read includes Meta data, which isn't relevant here; we just
            # want the AST for the content
            document_ast = pandoc.read(self.rendered_text, format=self.pandoc_format)
            children_ast = document_ast[1]
            return children_ast
        else:
            # Pandoc doesn't support reading "plain" text into it's AST (since
            # if it was just plain text, why would you need pandoc to parse it!)
            # That said, sometimes it is useful to create plain text from the
            # jinja rendering (e.g., when producing a site map or something
            # similar from Notion databases).
            return Plain([Str(self.rendered_text)])


notion_classes = {
    "page": JinjaRenderPage,
    "blocks": {
        "code": JinjaFencedCodeBlock,
    }
}



import pytest
from n2y import notion

from n2y.notion_mocks import (
    mock_formula_property_value, mock_person_user, mock_relation_value,
    mock_rich_text, mock_property_value, mock_rich_text_array,
    mock_rollup_property_value, mock_select_option, mock_user,
)


def process_property_value(notion_data):
    client = notion.Client('')
    property_value = client.wrap_notion_property_value(notion_data, None)
    return property_value.to_value('gfm')


def test_title():
    notion_data = mock_property_value('title', [mock_rich_text('text')])
    assert process_property_value(notion_data) == 'text'


def test_rich_text_empty():
    notion_data = mock_property_value('rich_text', [])
    assert process_property_value(notion_data) == ''


def test_rich_text_annotated():
    notion_data = mock_property_value('rich_text', mock_rich_text_array([
        ('Hello', ['bold']), (' ', []), ('Goodbye', 'italic')
    ]))
    assert process_property_value(notion_data) == '**Hello** *Goodbye*\n'


def test_number_empty():
    notion_data = mock_property_value('number', None)
    assert process_property_value(notion_data) is None


def test_number_non_empty():
    notion_data = mock_property_value('number', 3)
    assert process_property_value(notion_data) == 3


def test_select_empty():
    notion_data = mock_property_value('select', None)
    assert process_property_value(notion_data) is None


def test_select_non_empty():
    notion_data = mock_property_value('select', mock_select_option('A'))
    assert process_property_value(notion_data) == 'A'


def test_multi_select_empty():
    notion_data = mock_property_value('multi_select', [])
    assert process_property_value(notion_data) == []


def test_multi_select_non_empty():
    notion_data = mock_property_value('multi_select', [
        mock_select_option('A'),
        mock_select_option('B'),
    ])
    assert process_property_value(notion_data) == ['A', 'B']


def test_date_empty():
    notion_data = mock_property_value('date', None)
    assert process_property_value(notion_data) is None


def test_date_start_only():
    notion_data = mock_property_value('date', {'start': '2022-05-11'})
    assert process_property_value(notion_data) == '2022-05-11'


def test_date_start_and_end():
    notion_data = mock_property_value('date', {'start': '2022-05-11', 'end': '2022-05-12'})
    assert process_property_value(notion_data) == ['2022-05-11', '2022-05-12']


def test_people_empty():
    notion_data = mock_property_value('people', [])
    assert process_property_value(notion_data) == []


def test_people_minimal_user():
    notion_data = mock_property_value('people', [mock_user()])
    assert process_property_value(notion_data) == ['']


def test_people_person_user():
    notion_person = mock_person_user("Happy Man", 'happyman@gmail.com')
    notion_data = mock_property_value('people', [notion_person])
    assert process_property_value(notion_data) == ['Happy Man']


# TODO: add tests for bot users


def test_checkbox():
    notion_data = mock_property_value('checkbox', True)
    assert process_property_value(notion_data)


def test_url_empty():
    notion_data = mock_property_value('url', None)
    assert process_property_value(notion_data) is None


def test_url_non_empty():
    notion_data = mock_property_value('url', "https://innolitics.com")
    assert process_property_value(notion_data) == "https://innolitics.com"


def test_phone_number_empty():
    notion_data = mock_property_value('phone_number', None)
    assert process_property_value(notion_data) is None


def test_phone_number_non_empty():
    notion_data = mock_property_value('phone_number', "1-555-555-5555")
    assert process_property_value(notion_data) == "1-555-555-5555"


def test_email_empty():
    notion_data = mock_property_value('email', None)
    assert process_property_value(notion_data) is None


def test_email_non_empty():
    notion_data = mock_property_value('email', "info@innolitics.com")
    assert process_property_value(notion_data) == "info@innolitics.com"


def test_formula_number_empty():
    notion_data = mock_formula_property_value('number', None)
    assert process_property_value(notion_data) is None


def test_formula_number_non_empty():
    notion_data = mock_formula_property_value('number', 3)
    assert process_property_value(notion_data) == 3


def test_formula_date_empty():
    notion_data = mock_formula_property_value('date', None)
    assert process_property_value(notion_data) is None


def test_formula_date_start_only():
    notion_data = mock_formula_property_value('date', {'start': '2022-05-11'})
    assert process_property_value(notion_data) == '2022-05-11'


def test_relationship_empty():
    notion_data = mock_property_value('relation', [])
    assert process_property_value(notion_data) == []


def test_relationship_single():
    item = mock_relation_value()
    notion_data = mock_property_value('relation', [item])
    assert process_property_value(notion_data) == [item["id"]]


def test_relationship_double():
    item1 = mock_relation_value()
    item2 = mock_relation_value()
    notion_data = mock_property_value('relation', [item1, item2])
    assert process_property_value(notion_data) == [item1["id"], item2["id"]]


def test_rollup_date_empty():
    notion_data = mock_rollup_property_value('date', None)
    assert process_property_value(notion_data) is None


# TODO: Add some more tests for different types of rollups, including arrays of
# different types of values


@pytest.mark.xfail
def test_rollup_date_array():
    notion_data = mock_rollup_property_value('array', [
        {"start": "2022-05-11"},
        {"start": "2022-05-12"},
        {"start": "2022-05-13"},
    ])
    assert process_property_value(notion_data) == ['2022-05-11', '2022-05-12', '2022-05-12']



import logging

from n2y.utils import fromisoformat, sanitize_filename


logger = logging.getLogger(__name__)


class Database:
    def __init__(self, client, notion_data):
        logger.debug("Instantiating database")
        self.client = client

        self.notion_data = notion_data
        self.notion_id = notion_data['id']
        self.created_time = fromisoformat(notion_data['created_time'])
        self.created_by = client.wrap_notion_user(notion_data['created_by'])
        self.last_edited_time = fromisoformat(notion_data['last_edited_time'])
        self.last_edited_by = client.wrap_notion_user(notion_data['last_edited_by'])
        self.title = client.wrap_notion_rich_text_array(notion_data['title'])
        self.icon = notion_data['icon'] and client.wrap_notion_file(notion_data['icon'])
        self.cover = notion_data['cover'] and client.wrap_notion_file(notion_data['cover'])
        self.archived = notion_data['archived']
        self.schema = {
            k: client.wrap_notion_property(p)
            for k, p in notion_data['properties'].items()
        }
        self.notion_parent = notion_data['parent']
        self.notion_url = notion_data['url']
        self.archived = notion_data['archived']

        self._children = None
        self._filtered_children = {}

    @property
    def filename(self):
        return sanitize_filename(self.title.to_plain_text())

    @property
    def children(self):
        if self._children is None:
            self._children = self.client.get_database_pages(self.notion_id)
        return self._children

    def children_filtered(self, filter, sort=None):
        if filter is not None:
            tupled_filter = self._tuplize(filter)
            tupled_sort = self._tuplize(sort)
            if tupled_filter not in self._filtered_children:
                self._filtered_children[tupled_filter] = {}
            if tupled_sort not in self._filtered_children[tupled_filter]:
                self._filtered_children[tupled_filter][tupled_sort] = \
                    self.client.get_database_pages(self.notion_id, filter, sort)
            children = self._filtered_children[tupled_filter][tupled_sort]
        else:
            children = self.children
        for i, child in enumerate(children):
            if not self.client.page_class_is_in_use(child):
                children[i] = self.client._wrap_notion_page(child.notion_data)
        return children

    def _tuplize(self, item):
        if callable(getattr(item, "items", None)):
            return tuple([(key, self._tuplize(val)) for (key, val) in item.items()])
        elif hasattr(item, '__iter__') and type(item) is not str:
            return tuple([self._tuplize(i) for i in item])
        else:
            return (item)

    @property
    def parent(self):
        if self.notion_parent["type"] == "workspace":
            return None
        else:
            return self.client.get_page(self.notion_parent["page_id"])

    def to_pandoc(self):
        return self.block.to_pandoc()



from n2y.blocks import CalloutBlock


class RemoveCalloutBlock(CalloutBlock):
    """
    Completely removes callout blocks from the generated content.

    Useful, e.g., if you want callout blocks to contain help text that shouldn't
    be included in generated documents.
    """

    def __init__(self, client, notion_data, page, get_children=True):
        pass

    def to_pandoc(self):
        # It should be possible to return `pandoc.types.Null`, but that didn't
        # seem to work due to `TypeError: Object of type MetaType is not JSON serializable`
        # TODO: investigate why and simplify Block.children_to_pandoc if it can
        # be supported.
        return None


notion_classes = {
    "blocks": {
        "callout": RemoveCalloutBlock,
    }
}



from datetime import datetime
import uuid

from n2y.utils import strip_hyphens


def mock_id():
    return str(uuid.uuid4())


def mock_user(**kwargs):
    return {'object': 'user', 'id': mock_id(), **kwargs}


def mock_person_user(name, email):
    return mock_user(name=name, type="person", person={"email": email})


def mock_rich_text_array(text_blocks_descriptors):
    if isinstance(text_blocks_descriptors, str):
        return [mock_rich_text(text_blocks_descriptors, [])]
    else:
        return [mock_rich_text(*desc) for desc in text_blocks_descriptors]


def mock_rich_text(text, annotations=None, href=None, mention=None, link=None, equation=None):
    if annotations is None:
        annotations = []
    if mention:
        rich_text_type = 'mention'
        content = mention
    elif equation:
        rich_text_type = 'equation'
        content = equation
    else:
        rich_text_type = 'text'
        content = {'content': text, 'link': link}
    return {
        'type': rich_text_type,
        'annotations': mock_annotations(annotations),
        'plain_text': text,
        'href': href,
        rich_text_type: content,
    }


def mock_annotations(annotations=None):
    if annotations is None:
        annotations = []
    color_string = [a for a in annotations if 'color' in a]
    if color_string:
        color = color_string[0].replace('color:', '')
    else:
        color = 'default'
    return {
        'bold': True if 'bold' in annotations else False,
        'italic': True if 'italic' in annotations else False,
        'strikethrough': True if 'strikethrough' in annotations else False,
        'underline': True if 'underline' in annotations else False,
        'code': True if 'code' in annotations else False,
        'color': color
    }


def mock_user_mention():
    return {
        'type': 'user',
        'user': mock_user(),
    }


def mock_page_mention():
    return {
        'type': 'page',
        'page': {
            'id': mock_id(),
        },
    }


def mock_database_mention(id=None):
    return {
        'type': 'database',
        'database': {
            'id': id or mock_id(),
        },
    }


def mock_block(block_type, content, has_children=False, **kwargs):
    created_by = mock_user()
    created_time = datetime.now().isoformat()
    notion_id = mock_id()
    return {
        'id': notion_id,
        'url': f'#{strip_hyphens(notion_id)}',
        'created_time': created_time,
        'created_by': created_by,
        'last_edited_time': created_time,
        'last_edited_by': created_by,
        'object': 'block',
        'has_children': has_children,
        'archived': False,
        'type': block_type,
        block_type: content,
        **kwargs,
    }


def mock_paragraph_block(text_blocks_descriptors, **kwargs):
    return mock_block('paragraph', {
        'color': 'default',
        'rich_text': mock_rich_text_array(text_blocks_descriptors),
    }, **kwargs)


def mock_file(url):
    return {
        'url': url,
        'expiry_time': None
    }


def mock_property_value(property_value_type, content):
    return {
        'id': mock_id(),
        'type': property_value_type,
        property_value_type: content,
    }


def mock_rich_text_property_value(text_blocks_descriptors):
    return mock_property_value("rich_text", mock_rich_text_array(text_blocks_descriptors))


def mock_formula_property_value(formula_type, content):
    return mock_property_value("formula", {
        "type": formula_type,
        formula_type: content,
    })


def mock_rollup_property_value(rollup_type, content):
    return mock_property_value("rollup", {
        "type": rollup_type,
        "function": "function",
        rollup_type: content,
    })


def mock_select_option(name, **kwargs):
    return {
        'id': mock_id(),
        'name': name,
        'color': 'green',
        **kwargs,
    }


def mock_relation_value():
    return {"id": mock_id()}


def mock_page(title="Mock Page", extra_properties=None):
    if extra_properties is None:
        extra_properties = {}
    user = mock_user()
    created_time = datetime.now().isoformat()
    notion_id = mock_id()
    hyphenated_title = title.replace(" ", "-")
    return {
        'object': 'page',
        'id': notion_id,
        'created_time': created_time,
        'last_edited_time': created_time,
        'created_by': user,
        'last_edited_by': user,
        'cover': None,
        'icon': None,
        'parent': {'type': 'page_id', 'page_id': mock_id()},
        'archived': False,
        'properties': {
            'title': {
                'id': 'title',
                'type': 'title',
                'title': mock_rich_text_array(title)
            }, **extra_properties,
        },
        'url': f'https://www.notion.so/{hyphenated_title}-{notion_id}',
    }


def mock_database(title='Mock Database', extra_properties=None):
    if extra_properties is None:
        extra_properties = {}
    hyphenated_title = title.replace(" ", "-")
    created_time = datetime.now().isoformat()
    notion_id = mock_id()
    user = mock_user()
    return {
        'object': 'database',
        'id': notion_id,
        'cover': None,
        'icon': None,
        'created_time': created_time,
        'last_edited_time': created_time,
        'created_by': user,
        'last_edited_by': user,
        'title': mock_rich_text_array(title),
        'description': [],
        'is_inline': False,
        'archived': False,
        'properties': {
            'Name': {
                'id': 'title',  # all title properties have id 'title'
                'name': 'Name',
                'type': 'title',
                'title': {}
            }, **extra_properties,
        },
        'parent': {
            'type': 'page_id',
            'page_id': mock_id()
        },
        'url': f'https://www.notion.so/{hyphenated_title}-{notion_id}',
    }
